/* 
   Author: Joel W. Reed (joelreed@yahoo.com).  Copyright 1999.
   This file is distributed  under the Artistic License.
   See the file LICENSE for details.
*/

#include <stdio.h>
#include <math.h>
#include <string>
#include "nn-structs.h"

int parse_size, verbose, show_help, silent;
char outfname[256]; char network_fname[256]; char pattern_fname[256]; 
std::vector<std::string> l1_tags, ln_tags;
std::vector<int> nn_dims;

#define ERRMSG(msg) if (!silent) fprintf(stderr, msg);

fpnum_t activation(fpnum_t h)
{
	//const static double steepness = 0.5;
	// the below should really be 1/(1+exp(2*steepness*h))
	// but since i'm setting steepness to .5 (which works well for 0/1 range)
	// i've simplified the expression below for speed
	return 1.0/(1+exp(h));
}

fpnum_t compute_delta(fpnum_t obtained, fpnum_t delta)
{
	// again steepness set to .5 (see note in function activation())
	// which simplifies this equation
	return obtained * (1 - obtained)*delta;
}

unit::unit(const char* t, fpnum_t v) : _value(v), _tag(t)
{
}

weight::weight(fpnum_t v, unit* f, unit* s)  : _value(v), _first(f), _second(s)
{
}

network::network()
{
}

bool network::initialize(const std::vector<std::string>& l1_tags, const std::vector<std::string>& ln_tags, const std::vector<int>& dims)
{
	// create unit layer 1 from tag list
	{
		unit_layer l;
		for (std::vector<std::string>::const_iterator i = l1_tags.begin(); i != l1_tags.end(); ++i)
		{
			l.push_back(unit(i->c_str()));
		}       
		_unit_layers.push_back(l);
	}

	// create unit layers 2->N based on requested dimensions
	for (std::vector<int>::const_iterator i = dims.begin(); i != dims.end(); ++i)
	{
		unit_layer t(*i);
		_unit_layers.push_back(t);
	}

	// create unit layer 'N' from tag list
	{
		unit_layer l;
		for (std::vector<std::string>::const_iterator i = ln_tags.begin(); i != ln_tags.end(); ++i)
		{
			l.push_back(unit(i->c_str()));
		}       
		_unit_layers.push_back(l);
	}

	// now connect the layers with the weights
	for (unit_layers_t::iterator i = _unit_layers.begin(), j = i+1; j != _unit_layers.end(); ++i, ++j)
	{
		weight_layer wl;

		for (unit_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			double multiplier = 1/sqrt(i->size());
			for (unit_layer::iterator m = j->begin(); m != j->end(); ++m)
			{
				double wv = (rand()/((double)RAND_MAX))*multiplier;
				weight w(wv, &(*k), &(*m));
				wl.push_back(w);
			}
		}

		_weight_layers.push_back(wl);
	}
}

bool network::load(const char* fn)
{
	// pattern can be read from stdin, not a saved network. 
	FILE* file = fopen(fn, "r");

	while (!feof(file))
	{
		char type[256]; 
		if (0==fgets(type, sizeof(type), file))
			break;

		if (strstr(type, "unit "))
		{
			unit_layer layer;
			int cnt = atoi(type+5); // get to cnt. yes i know this is ugly
			
			while (!feof(file) && cnt > 0)
			{
				char tag[MAX_TAG_SIZE];
				fpnum_t value; fscanf(file, "%s %f\n", tag, &value);
				layer.push_back(unit(tag, value));
				--cnt;
			}

			_unit_layers.push_back(layer);
		}
		else if (strstr(type, "weight "))
		{
			weight_layer layer;
			int cnt = atoi(type+7);
			
			while (!feof(file) && cnt > 0)
			{
				char num[256]; fgets(num, sizeof(num), file);
				fpnum_t value = atof(num);
				layer.push_back(weight(value));
				--cnt;
			}
			
			_weight_layers.push_back(layer);
		}
	}

	// now connect the layers with the weights
	weight_layers_t::iterator wli = _weight_layers.begin();
	for (unit_layers_t::iterator i = _unit_layers.begin(), j = i+1; j != _unit_layers.end(); ++i, ++j, ++wli)
	{
		weight_layer::iterator wi = wli->begin();
		for (unit_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			for (unit_layer::iterator m = j->begin(); m != j->end(); ++m, ++wi)
			{
				wi->_first = &(*k);
				wi->_second = &(*m);
			}
		}
	}

	fclose(file);
	return true;
}

bool network::save(const char* fn) const
{
	FILE* file = ((0==fn)||('-' == *fn))? stdout : fopen(fn, "w");

	// now connect the layers with weights
	int cnt = 0;
	for (unit_layers_t::const_iterator i = _unit_layers.begin(); i != _unit_layers.end(); ++i, ++cnt)
	{
		fprintf(file, "unit %d\n", i->size());
		for (unit_layer::const_iterator k = i->begin(); k != i->end(); ++k)
		{
			fprintf(file, "%s %f\n", k->_tag.c_str(), k->_value);
		}
	}

	cnt = 0;
	for (weight_layers_t::const_iterator i = _weight_layers.begin(); i != _weight_layers.end(); ++i, ++cnt)
	{
		fprintf(file, "weight %d\n", i->size());
		for (weight_layer::const_iterator k = i->begin(); k != i->end(); ++k)
		{
			fprintf(file, "%f\n", k->_value);
		}
	}

	if ((0!=fn)&&('-' != *fn)) fclose(file);
	return true;
}

bool network::process(const unit_layer& inputs)
{
	// get reference to input layer of nn
	unit_layer& t = *(_unit_layers.begin()); 

	for (unit_layer::const_iterator i = inputs.begin(); i != inputs.end(); ++i)
	{
 		bool unused = true;
		for (unit_layer::iterator j = t.begin(); j != t.end(); ++j)
		{
 			if (j->_tag == i->_tag)
			{
				j->_value = j->_value; 
				unused = false;
				break;
			}
		}       
		
		if ((unused)&&(!silent)) fprintf(stderr, "warning: network input layer does not have an unit for '%s'\n", i->_tag.c_str()); 
	}

	// propogate the signal thru the network...
	for (weight_layers_t::iterator i = _weight_layers.begin(); i != _weight_layers.end(); ++i)
	{
		// first sum the weights
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			k->_second->_value += (k->_first->_value * k->_value);  
		}

		// then apply the activation function
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			k->_second->_value = activation(-k->_second->_value);
		}
	}

	return true;
}

bool network::learn(const unit_layer& desired_outputs)
{
	// get reference to output layer of nn
	unit_layer& t = *(_unit_layers.rbegin()); 
	
	unit_layer::const_iterator doi = desired_outputs.begin();
	for (unit_layer::iterator i = t.begin(); i != t.end(); ++i, ++doi)
	{
		i->_delta = compute_delta(i->_value, doi->_value - i->_value);
	}

	// propogate the signal thru the network...
	for (weight_layers_t::reverse_iterator i = _weight_layers.rbegin(); i != _weight_layers.rend(); ++i)
	{
		// first sum the weights
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			k->_first->_delta += (k->_second->_delta * k->_value);  
		}

		// then apply the delta function
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			k->_first->_delta = compute_delta(k->_first->_value, k->_first->_delta);
		}
	}

	// now adjust our weights based on the deltas
	unit_layers_t::const_iterator ui = _unit_layers.begin();
	for (weight_layers_t::iterator i = _weight_layers.begin(); i != _weight_layers.end(); ++i, ++ui)
	{
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			//TODO: change 1/unit_cnt to some other adaptive pattern (page 124..)
			k->_value = k->_first->_value * k->_first->_delta * (1.0/ui->size());
		}
	}

	return true;
}

bool pattern::load(const char* fn)
{
	// load input pattern...
	FILE* file = (*fn=='-')? stdin: fopen(fn, "r");

	while (!feof(file))
	{
		char type[256]; 
		if (0==fgets(type, sizeof(type), file))
			break;

		if (strstr(type, "input "))
		{
			int cnt = atoi(type+6); // get to cnt. yes i know this is ugly
			
			while (!feof(file) && cnt > 0)
			{
				char tag[MAX_TAG_SIZE];
				fpnum_t value; fscanf(file, "%s %f\n", tag, &value);
				_input.push_back(unit(tag, value));
				--cnt;
			}
		}
		else if (strstr(type, "output "))
		{
			int cnt = atoi(type+7);
			
			while (!feof(file) && cnt > 0)
			{
				char tag[MAX_TAG_SIZE];
				fpnum_t value; fscanf(file, "%s %f\n", tag, &value);
				_output.push_back(unit(tag, value));
				--cnt;
			}
		}
	}

	if (_input.size()==0)
	{
		ERRMSG("Error: No input data in learning pattern.\n");
		return false;
	}

	if (_output.size()==0)
	{
		ERRMSG("Error: No desired output data in learning pattern.\n");
		return false;
	}

	if ('-' != *fn) fclose(file);
	return true;
}

///////////////////////////////////////////////////////////////////
#include "argtable.h"

int parse_args(int argc, char* argv[])
{
	char cl1_tags[4096], cln_tags[4096];
	char cnn_dims[256];
	static arg_rec argtable[] =
	{
		/*-tag---------name---------type-----storage---default---descrip --*/
		{ "-f ",  NULL, arg_str,  cl1_tags,  "",   "\t is a comma separated list of the first layer's\n\t\t input keys. Also used to determine layer's node count." },
		{ "-l ",  NULL, arg_str,  cln_tags,  "",   "\t is a comma separated list of the last layer's\n\t\t output keys. Also used to determine the layer's node count." },
		{ "-h ",  NULL, arg_str,  cnn_dims,  "",   "\t is a comma separated list of node counts for the hidden \n\t\t layers 2->(N-1). No spaces allowed!" },
		{ "-v ",  NULL, arg_lit,  &verbose, "0",     "\t\t specifies verbose output" },
		{ "-s ",  NULL, arg_lit,  &silent, "0",     "\t\t supress all error message & warnings." },
		{ "-h ",  NULL,        arg_lit,  &show_help, "0",     "\t\t show command line help." },
		{ "-p ", "<pattern>", arg_str,  pattern_fname,  "-",    "\t is the pattern input file, use \"-\" for stdin (default=\"-\")" },
		{ "-n ", "<network>", arg_str, network_fname,  "",    "\t is filename of a saved network." },
		{ "-o ",  "<outfile>", arg_str,  outfname, "-",     "\t is the output file, use \"-\" for stdout (default=\"-\")" },
	};

	static const size_t narg = sizeof(argtable)/sizeof(arg_rec);

	/*-- scan command line arguments from argv[] as per argtable[] specs --*/
	char cmdline[200], errmsg[200], errmark[200]; 
	if (!arg_scanargv(argc,argv,argtable,narg,cmdline,errmsg,errmark))
	{
		/*-- scan error occurred, print error message and exit --*/
		printf("ERROR: %s\n", cmdline);
		printf("       %s %s\n", errmark, errmsg);
		exit(1);
	}

	if ((argc==1)||(show_help))
	{
		/*-- no command line arguments, so display program usage and exit. --*/
		printf("'nn' is a scriptable neural network, a component of NTSAP.\n");
		printf("Usage: %s %s\n",argv[0],arg_syntax(argtable,narg));
		printf("%s\n",arg_glossary(argtable,narg,"   "));
		puts("When initially creating a neural network, you must specify");
		puts("the first layer tags using -f, the last layer's tags using -l, ");
		puts("and the node sizes for layers 2->N using -n.\n");
		exit(0);
	}

	// now parse up create_list if need be
	if (*cnn_dims!='\0')
	{
		std::string t;
		char* i = cnn_dims;
		
		while (i != 0)
		{
			char* comma = strchr(i, ',');
			if (comma) t = std::string(i, comma);
			else t = std::string(i);
			
			nn_dims.push_back(atoi(t.c_str()));
			i = (comma)? comma+1 : 0;
		}
	}
	
	// now parse layer one tags
	if (*cl1_tags != '\0')
	{
		std::string t;
		char* i = cl1_tags;
		
		while (i != 0)
		{
			char* comma = strchr(i, ',');
			if (comma) t = std::string(i, comma);
			else t = std::string(i);
			
			l1_tags.push_back(t);
			i = (comma)? comma+1 : 0;
		}		
	}

	// now parse layer 'N' tags
	if (*cln_tags != '\0')
	{
		std::string t;
		char* i = cln_tags;
		
		while (i != 0)
		{
			char* comma = strchr(i, ',');
			if (comma) t = std::string(i, comma);
			else t = std::string(i);
			
			ln_tags.push_back(t);
			i = (comma)? comma+1 : 0;
		}		
	}

	if ((nn_dims.size()==0)&&(*network_fname=='\0'))
	{
		puts("Please specify either a new network's parameters (-f -l -h)");
		puts("  or a the filename of a saved neural network.");
		exit(1);
	}

	return 0;
}

///////////////////////////////////////////////////////////////////

int main (int argc, char* argv[])
{
	parse_args(argc, argv);

	// seed random number generator
	srand(time(0));

	// for now if a network is being created from scratch we don't also do a process/learn. could change later
	network nn;
	if (nn_dims.size()>0)
	{
		nn.initialize(l1_tags, ln_tags, nn_dims);
		nn.save();
		exit(0);
	}
	else
	{
		if (!nn.load(network_fname)) 
		{
			ERRMSG("Error: Failed to load neural network. Please confirm the file is appropriately formatted.\n");
			return 1;
		}

		pattern learning_pattern; 
		if (!learning_pattern.load(pattern_fname))
		{
			ERRMSG("Error: Failed to load learning pattern. Please confirm the file is appropriately formatted.\n");
			return 1;
		}

		nn.process(learning_pattern._input); // compute output
		nn.learn(learning_pattern._output); // adjust weights based on deviance

		nn.save();
	}
	
	return 0;
}
